import asyncio
import logging
import signal

import hypercorn.asyncio
import starlette


async def serve_http(app: starlette.applications.Starlette, port: int = 8000) -> None:
    config = hypercorn.Config()
    config.bind = [f"0.0.0.0:{int(port)}"]
    config.accesslog = logging.getLogger("hypercorn.access")
    config.access_log_format = '[%(s)s] %(m)s %(U)s?%(q)s [%(L)ss] %(b)s "%(h)s"'
    config.errorlog = logging.getLogger("hypercorn.error")

    shutdown_trigger = None
    has_custom_handlers = any(
        signal.getsignal(signum) not in {signal.default_int_handler, signal.SIG_DFL}
        for signum in (signal.SIGINT, signal.SIGTERM)
    )
    if has_custom_handlers:
        # prevent hypercorn from catching `SIGTERM` and `SIGINT` if
        # other signal handlers are already installed elsewhere.
        shutdown_trigger = asyncio.Event().wait

    await hypercorn.asyncio.serve(app, config, shutdown_trigger=shutdown_trigger)



import atexit
import contextlib
import logging
import os
import re
import typing

from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.instrumentation.httpx import HTTPXClientInstrumentor
from opentelemetry.instrumentation.pymongo import PymongoInstrumentor
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import ReadableSpan, TracerProvider
from opentelemetry.sdk.trace.export import (
    BatchSpanProcessor,
    ConsoleSpanExporter,
    SpanExporter,
    SpanExportResult,
)

from delphai_utils.logging import register_custom_field

SERVICE_NAME_RE = re.compile(r"[^a-z0-9-]+")


class UniqueMongoCommandsSpanExporter(SpanExporter):
    IGNORE_STATEMENTS = ["saslContinue", "getMore"]

    def __init__(self):
        self.logger = logging.getLogger("UniqueMongoCommandsSpanExporter")
        self.unique_statements = {}

        atexit.register(self.log_unique_statements)

    def export(self, spans: typing.Sequence[ReadableSpan]) -> SpanExportResult:
        server_span_names = {}
        mongo_statements_per_trace = {}

        for span in spans:
            trace_id = span.context.trace_id
            if not trace_id:
                continue

            if span.kind == trace.SpanKind.SERVER:
                server_span_names[trace_id] = span.name
                continue

            if span.kind != trace.SpanKind.CLIENT:
                continue

            scope_name = span.instrumentation_scope.name
            if scope_name != "opentelemetry.instrumentation.pymongo":
                continue

            if span.attributes.get("db.statement") in self.IGNORE_STATEMENTS:
                continue

            mongo_statements_per_trace.setdefault(trace_id, set()).add(
                frozenset(
                    {
                        "database": span.attributes.get("db.name"),
                        "collection": span.attributes.get("db.mongodb.collection"),
                        "statement": span.attributes.get("db.statement"),
                    }.items()
                )
            )

        updated = False
        for trace_id, mongo_requests in mongo_statements_per_trace.items():
            server_request_name = server_span_names.get(trace_id)
            if not server_request_name:
                continue

            unique_statements = self.unique_statements.setdefault(
                server_request_name, set()
            )

            if mongo_requests <= unique_statements:
                continue

            updated = True
            unique_statements |= mongo_requests

        if updated:
            self.log_unique_statements()

        return SpanExportResult.SUCCESS

    def log_unique_statements(self):
        for name, queries in sorted(self.unique_statements.items()):
            for query in queries:
                query = dict(query)
                query = f"{query.get('database')}.{query.get('collection')}: {query.get('statement')}"
                self.logger.info("%s => %s", name, query)


def instrument_mongo():
    def _pymongo_request_hook(span, event):
        if not span.is_recording():
            return

        collection = span.attributes.get("db.mongodb.collection") or "*"
        span_name = f"mongo:{event.database_name}.{collection}.{event.command_name}"

        span.update_name(span_name)

    # Applied once per process
    # Affects only instances created after patch
    PymongoInstrumentor().instrument(request_hook=_pymongo_request_hook)


def configure_tracing(service_name):
    use_exporter_console = os.getenv("OTEL_EXPORTER_CONSOLE")
    use_exporter_otlp = os.getenv("OTEL_EXPORTER_OTLP_ENDPOINT") or os.getenv(
        "OTEL_EXPORTER_OTLP_TRACES_ENDPOINT"
    )

    if not use_exporter_console and not use_exporter_otlp:
        return

    # Allow to overwrite service name
    service_name = SERVICE_NAME_RE.sub("-", service_name)
    service_name = os.getenv("OTEL_SERVICE_NAME", service_name).strip().lower()
    resource = Resource.create(attributes={SERVICE_NAME: service_name})

    tracer_provider = TracerProvider(resource=resource)

    if use_exporter_console:
        tracer_provider.add_span_processor(BatchSpanProcessor(ConsoleSpanExporter()))

    if use_exporter_otlp:
        tracer_provider.add_span_processor(BatchSpanProcessor(OTLPSpanExporter()))

    with contextlib.suppress(ImportError):
        import httpx  # noqa: F401

        _instrument_httpx()

    with contextlib.suppress(ImportError):
        import pymongo  # noqa: F401

        tracer_provider.add_span_processor(
            BatchSpanProcessor(UniqueMongoCommandsSpanExporter())
        )

    trace.set_tracer_provider(tracer_provider)

    def get_trace_id():
        span_context = trace.get_current_span().get_span_context()
        if span_context.trace_id:
            trace_id = span_context.trace_id
            span_id = span_context.span_id
            return f"{trace_id:016x}:{span_id:016x}"

    register_custom_field("trace_id", get_trace_id)


def _instrument_httpx():
    def request_hook(span, request):
        span.update_name(f"httpx:{span.name}")

    async def async_request_hook(span, request):
        span.update_name(f"httpx:{span.name}")

    # Applied once per process
    # Patches class methods, affects new and existing instances
    HTTPXClientInstrumentor().instrument(
        request_hook=request_hook,
        async_request_hook=async_request_hook,
    )



import re
import unicodedata

from urllib.parse import quote, unquote, urlencode

import can_ada


SCHEME_RE = re.compile(r"^[a-zA-Z][a-zA-Z0-9+.-]*://")


def normalize_url(url: str, default_scheme: str = "https") -> str:
    """
    Transforms the raw_url with following rules:
    * Lowercase hostname
    * Lowercase schema
    * Remove default ports
    * Uppercase the percent-encoding
    * Remove trailing slashes in path
    * Remove trailing dots at end of hostnames
    * Sort query parameters:
      * sort names and then values as case sensitive strings
      * keep order of arrays ( example: "a[]=y&a[]=x" )
    * Decode international characters and UTF8
    * Normalize UTF8 with "NFKC" rules
    * percent-encode reserved characters if needed:
      * reserved characters are defined in https://datatracker.ietf.org/doc/html/rfc3986#section-2.2
      * these characters are encoded if they appear in places other than defined in https://www.rfc-editor.org/rfc/rfc3986.html
      * mainly, the path or query parts of URL are affected

    Uses can-ada url parser (https://pypi.org/project/can-ada/).
    It is compliant with the standard https://url.spec.whatwg.org/
    and  ~4x faster than urllib
    """
    url = unicodedata.normalize("NFKC", url)

    if not SCHEME_RE.match(url):
        if not default_scheme:
            raise ValueError("default_scheme must not be empty")
        url = f"{default_scheme}://{url.lstrip('/')}"

    parsed_url = can_ada.parse(url)
    assert parsed_url.hostname

    # Ada (can_ada) validates values in setters and silently ignores invalid
    # Below we ensure that the value was set and not silently ignored
    hostname = parsed_url.hostname.rstrip(".")
    parsed_url.hostname = hostname
    assert parsed_url.hostname == hostname

    pathname = quote(unquote(parsed_url.pathname)).rstrip("/") or "/"
    parsed_url.pathname = pathname
    assert parsed_url.pathname == pathname

    search = _normalize_url_query(parsed_url.search)
    parsed_url.search = search
    assert parsed_url.search == search

    return str(parsed_url)


def _normalize_url_query_sort_key(key_value):
    key, value = key_value
    if "[]" in key:
        return f"{key}="
    else:
        return f"{key}={value}"


def _normalize_url_query(raw_url_query: str) -> str:
    url_query = raw_url_query.lstrip("?")
    if not url_query:
        return ""
    filtered_query = [
        (key, value)
        for key, value in list(can_ada.URLSearchParams(url_query))
        if not key.startswith("utm_")
    ]
    sorted_query = sorted(filtered_query, key=_normalize_url_query_sort_key)
    if not sorted_query:
        return ""

    return "?" + urlencode(sorted_query)


def clean_url(url, keep_www=False):
    """
    Format and clean an url to be saved or checked.
    Args:
        url: url to be formatted
        keep_www: keep the 'www' part of the url
    Returns: formatted url
    """

    url = url.strip()
    url = url.replace("https://", "").replace("http://", "").rstrip("/")
    if not keep_www:
        url = url.replace("www.", "")
    split_url = url.split("/")
    split_url[0] = split_url[0].lower()
    return "/".join(split_url)


def get_clean_domain(url):
    """
    Format and clean an url and returns domain.
    Args:
        url: url to be formatted
    Returns: formatted domain
    """

    return clean_url(url).split("/")[0]
